import codecs
import datetime
import io
import json as json_
from collections import abc
from urllib.parse import urlsplit, urlunsplit, urlencode, parse_qsl, quote
from io import BytesIO
from typing import *

import chardet

from pycurl_requests.cookies import RequestsCookieJar
from pycurl_requests import exceptions
from pycurl_requests import structures

DEFAULT_REDIRECT_LIMIT = 30


class Request:
    def __init__(self,
                 method=None,
                 url=None,
                 headers=None,
                 files=None,
                 data=None,
                 params=None,
                 auth=None,
                 cookies=None,
                 hooks=None,
                 json=None):

        self.method = method
        self.url = url
        self.headers = headers
        self.files = files
        self.data = data
        self.json = json
        self.params = params
        self.auth = auth
        self.cookies = cookies
        self.hooks = hooks

    def deregister_hook(self, event, hook):
        raise NotImplementedError

    def prepare(self):
        prepared = PreparedRequest()
        prepared.prepare(
            method=self.method,
            url=self.url,
            headers=self.headers,
            files=self.files,
            data=self.data,
            params=self.params,
            auth=self.auth,
            cookies=self.cookies,
            hooks=self.hooks,
            json=self.json)

        return prepared

    def register_hook(self, event, hook):
        raise NotImplementedError


class Response:
    def __init__(self):
        self.request = None  # type: Optional[Request]
        self.elapsed = None  # type: Optional[datetime.timedelta]
        self.status_code = None  # type: Optional[int]
        self.reason = None  # type: Optional[str]
        self.headers = None  # type: Optional[structures.CaseInsensitiveDict]
        self.encoding = None  # type: Optional[str]
        self.url = None  # type: Optional[str]
        self.raw = None  # type: Optional[BytesIO]

    @property
    def apparent_encoding(self):
        return chardet.detect(self.content)['encoding']

    def close(self):
        # Not implemented
        pass

    @property
    def content(self):
        return self.raw.getvalue()

    @property
    def cookies(self):
        return NotImplemented

    @property
    def history(self):
        return NotImplemented

    @property
    def is_permanent_redirect(self):
        # Moved Permanently (HTTP 301) or Permanent Redirect (HTTP 308)
        return self.status_code in {301, 308}

    @property
    def is_redirect(self):
        return self.status_code in {301, 302, 303, 307, 308}

    def iter_content(self, chunk_size=1, decode_unicode=False):
        chunk_size = chunk_size or -1
        decoder = codecs.getincrementaldecoder(self.encoding)('replace') if self.encoding and decode_unicode else None
        for chunk in iter(lambda: self.raw.read1(chunk_size), b''):
            if decoder:
                yield decoder.decode(chunk)
            else:
                yield chunk

        if decoder:
            # Make sure we finalize the decoder (may yield replacement character)
            tail = decoder.decode(b'', True)
            if tail:
                yield tail

    def iter_lines(self, chunk_size=512, decode_unicode=False, delimiter=None):
        leftover = None
        for chunk in self.iter_content(chunk_size, decode_unicode=decode_unicode):
            if leftover:
                chunk = leftover + chunk

            if delimiter is not None:
                parts = chunk.split(delimiter)
            else:
                parts = chunk.splitlines()

            # FIXME: This logic doesn't work for CR-only line endings
            if chr(ord(chunk[-1])) == '\n':
                yield from parts
                leftover = None
            else:
                # This may be a partial line, so add to the next chunk
                yield from parts[:-1]
                leftover = parts[-1]

        if leftover is not None:
            yield leftover

    def json(self, **kwargs):
        return json_.loads(self.content, **kwargs)

    @property
    def links(self):
        return NotImplemented

    @property
    def next(self):
        return NotImplemented

    @property
    def ok(self):
        return self.status_code < 400

    def raise_for_status(self):
        if 400 <= self.status_code < 500:
            raise exceptions.HTTPError('{s.status_code} Client Error: {s.reason} for url: {s.url}'.format(s=self),
                                       response=self)

        if 500 <= self.status_code < 600:
            raise exceptions.HTTPError('{s.status_code} Client Error: {s.reason} for url: {s.url}'.format(s=self),
                                       response=self)

    @property
    def text(self):
        return self.content.decode(self.encoding or 'ISO-8859-1')


class PreparedRequest:
    def __init__(self):
        self.method = None
        self.url = None
        self.headers = None
        self.body = None
        self.hooks = None

    @property
    def path_url(self):
        return urlsplit(self.url).path

    def prepare(self,
                method=None,
                url=None,
                headers=None,
                files=None,
                data=None,
                params=None,
                auth=None,
                cookies=None,
                hooks=None,
                json=None):

        self.prepare_method(method)
        self.prepare_url(url, params)
        self.prepare_headers(headers)
        self.prepare_cookies(cookies)
        self.prepare_body(data, files, json)
        self.prepare_auth(auth, url)
        self.prepare_hooks(hooks)

    def prepare_method(self, method):
        self.method = method.upper() if method else None

    def prepare_url(self, url, params):
        if isinstance(url, bytes):
            url = url.decode('iso-8859-1')

        url = url.strip()

        # Leave non-HTTP schemes as-is
        if ':' in url and not url.lower().startswith('http'):
            self.url = url
            return

        parts = urlsplit(url)
        path = quote(parts.path) if parts.path else '/'

        if not params:
            query = parts.query
        else:
            if isinstance(params, (str, bytes)):
                params = parse_qsl(params)
            if isinstance(params, abc.Mapping):
                params = list(params.items())
            else:
                params = list(params)

            query = urlencode(parse_qsl(parts.query) + params, doseq=True)

        self.url = urlunsplit(parts[:2] + (path, query) + parts[4:])

    def prepare_headers(self, headers):
        # NOTE: Only user-defined headers, not those set by libcurl
        headers = headers or structures.CaseInsensitiveDict()

        # Filter out headers with None value
        header_names = headers.keys()
        for name in header_names:
            if headers[name] is None:
                del headers[name]

        self.headers = headers

    def prepare_cookies(self, cookies):
        # Cookies can only be set if there is no existing `Cookie` header
        if 'Cookie' in self.headers or cookies is None:
            return

        cookiejar = RequestsCookieJar()
        cookiejar.update(cookies)

        value = '; '.join(('{}={}'.format(n, v) for n, v in cookiejar.iteritems()))

        self.headers['Cookie'] = value

    def prepare_content_length(self, body):
        content_length = None

        if body is None:
            if self.method not in ('GET', 'HEAD'):
                content_length = 0
        elif isinstance(body, bytes):
            content_length = len(body)
        elif isinstance(body, str):
            content_length = len(body.encode('iso-8859-1'))
        elif getattr(body, 'seekable', False):
            content_length = body.seek(0, io.SEEK_END)
            body.seek(0)

        if content_length is not None:
            self.headers['Content-Length'] = str(content_length)

    def prepare_body(self, data, files, json=None):
        body = None

        if files is not None:
            raise NotImplementedError
        elif data is not None:
            if isinstance(data, (io.RawIOBase, io.BufferedReader)):
                # It's a file-like object, so can be sent directly
                body = data
            elif isinstance(data, (abc.Mapping, list, tuple)):
                self._set_header_default('Content-Type', 'application/x-www-form-urlencoded')
                body = urlencode(data)
            else:
                # Assume it's something bytes-compatible
                body = data
        elif json is not None:
            self._set_header_default('Content-Type', 'application/json')
            body = json_.dumps(json, ensure_ascii=True).encode('ascii')

        if 'Content-Length' not in self.headers:
            self.prepare_content_length(body)

        self.body = body

    def _set_header_default(self, key, default):
        """Set header `key` to `default` if not already set"""
        if key not in self.headers:
            self.headers[key] = default

    def prepare_auth(self, auth, url=''):
        # FIXME: Not implemented
        pass

    def prepare_hooks(self, hooks):
        # FIXME: Not implemented
        pass



# This file is taken from Requests v2.22.0 with the following modifications:
#   - Replaced RequestException with custom implementation for PycURL-Requests
#   - Removed dependency on `urllib3.exceptions`
#   - Removed UTF-8 coding header comment
#
# It was originally released under the following licence:
# ```
# Copyright 2019 Kenneth Reitz
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        https://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
# ```

"""
requests.exceptions
~~~~~~~~~~~~~~~~~~~

This module contains the set of Requests' exceptions.
"""

import pycurl


class RequestException(IOError):
    """There was an ambiguous exception that occurred while handling your request."""

    @classmethod
    def from_pycurl_error(cls, error: pycurl.error, **kwargs):
        """Create a RequestException (or subclass) from PycURL error."""
        code, message = error.args[:2]
        msg = '{} (cURL error: {})'.format(message, code)

        exception = PYCURL_ERROR_MAPPING.get(code, cls)
        return exception(msg, curl_message=message, curl_code=code, **kwargs)

    def __init__(self, *args, curl_message=None, curl_code=None, request=None, response=None):
        self.curl_message = curl_message
        self.curl_code = curl_code
        self.request = request
        self.response = response

        if self.response is not None and not self.request and hasattr(self.response, 'request'):
            self.request = self.response.request

        super().__init__(*args)


class HTTPError(RequestException):
    """An HTTP error occurred."""


class ConnectionError(RequestException):
    """A Connection error occurred."""


class ProxyError(ConnectionError):
    """A proxy error occurred."""


class SSLError(ConnectionError):
    """An SSL error occurred."""


class Timeout(RequestException):
    """The request timed out.

    Catching this error will catch both
    :exc:`~requests.exceptions.ConnectTimeout` and
    :exc:`~requests.exceptions.ReadTimeout` errors.
    """


class ConnectTimeout(ConnectionError, Timeout):
    """The request timed out while trying to connect to the remote server.

    Requests that produced this error are safe to retry.
    """


class ReadTimeout(Timeout):
    """The server did not send any data in the allotted amount of time."""


class URLRequired(RequestException):
    """A valid URL is required to make a request."""


class TooManyRedirects(RequestException):
    """Too many redirects."""


class MissingSchema(RequestException, ValueError):
    """The URL schema (e.g. http or https) is missing."""


class InvalidSchema(RequestException, ValueError):
    """See defaults.py for valid schemas."""


class InvalidURL(RequestException, ValueError):
    """The URL provided was somehow invalid."""


class InvalidHeader(RequestException, ValueError):
    """The header value provided was somehow invalid."""


class InvalidProxyURL(InvalidURL):
    """The proxy URL provided is invalid."""


class ChunkedEncodingError(RequestException):
    """The server declared chunked encoding but sent an invalid chunk."""


class ContentDecodingError(RequestException):
    """Failed to decode response content"""


class StreamConsumedError(RequestException, TypeError):
    """The content for this response was already consumed"""


class RetryError(RequestException):
    """Custom retries logic failed"""


class UnrewindableBodyError(RequestException):
    """Requests encountered an error when trying to rewind a body"""

# Warnings


class RequestsWarning(Warning):
    """Base warning for Requests."""
    pass


class FileModeWarning(RequestsWarning, DeprecationWarning):
    """A file was opened in text mode, but Requests determined its binary length."""
    pass


class RequestsDependencyWarning(RequestsWarning):
    """An imported dependency doesn't match the expected version range."""
    pass


#: Mapping of PycURL error codes to Request exceptions
PYCURL_ERROR_MAPPING = {
    pycurl.E_UNSUPPORTED_PROTOCOL: ConnectionError,
    pycurl.E_URL_MALFORMAT: InvalidURL,
    pycurl.E_COULDNT_RESOLVE_PROXY: ProxyError,
    pycurl.E_COULDNT_RESOLVE_HOST: ConnectionError,
    pycurl.E_COULDNT_CONNECT: ConnectionError,
    pycurl.E_OPERATION_TIMEDOUT: Timeout,
    pycurl.E_SSL_CONNECT_ERROR: SSLError,
    pycurl.E_INTERFACE_FAILED: ConnectionError,
    pycurl.E_TOO_MANY_REDIRECTS: TooManyRedirects,
    pycurl.E_GOT_NOTHING: ConnectionError,
    pycurl.E_PEER_FAILED_VERIFICATION: SSLError,
    pycurl.E_SSL_CACERT: SSLError,
    pycurl.E_SSL_ISSUER_ERROR: SSLError,
    pycurl.E_SSL_PINNEDPUBKEYNOTMATCH: SSLError,
    pycurl.E_SSL_INVALIDCERTSTATUS: SSLError,
}



import datetime
import sys

import pytest

from pycurl_requests import cookies
from pycurl_requests import requests
from pycurl_requests.tests.utils import *  # Used for fixtures

if IS_PYCURL_REQUESTS:
    from pycurl_requests.structures import CaseInsensitiveDict
else:
    from requests.structures import CaseInsensitiveDict

try:
    from urllib3.util.timeout import Timeout
except ImportError:
    def Timeout(total=None, connect=None, read=None):
        return NotImplemented


def test_get(http_server):
    response = requests.get(http_server.base_url + '/hello')
    assert isinstance(response, requests.Response)
    response.raise_for_status()

    assert http_server.last_url.path == '/hello'

    assert isinstance(response.request, requests.PreparedRequest)
    assert response.request.method == 'GET'
    assert response.request.url == http_server.base_url + '/hello'
    assert response.request.path_url == '/hello'
    assert isinstance(response.request.headers, CaseInsensitiveDict)
    assert response.request.body is None

    assert response.elapsed > datetime.timedelta(0)
    assert response.url.endswith('/hello')
    assert response.status_code == 200
    assert response.reason == 'OK'
    assert response.headers['content-type'] == 'text/html; charset=UTF-8'
    assert response.headers['content-length'] == str(len(b'Hello\nWorld\n'))
    assert response.encoding.lower() == 'utf-8'
    assert response.apparent_encoding == 'ascii'
    assert response.content == b'Hello\nWorld\n'
    assert response.text == 'Hello\nWorld\n'


def test_get_notfound(http_server):
    response = requests.get(http_server.base_url + '/not_found')
    with pytest.raises(requests.HTTPError) as err:
        response.raise_for_status()
    assert err.value.response.status_code == 404


def test_get_params(http_server):
    response = requests.get(http_server.base_url + '/params', params={'q': 'foo', 'r': 'bar'})
    response.raise_for_status()

    assert response.text == 'q: foo\nr: bar'


@pytest.mark.parametrize('sequence_type', [list, tuple])
def test_get_params_list(http_server, sequence_type):
    response = requests.get(http_server.base_url + '/params', params=sequence_type([('q', 'foo'), ('r', 'bar'), ('q', 'baz')]))
    response.raise_for_status()

    assert response.text == 'q: foo\nr: bar\nq: baz'


def test_get_headers(http_server):
    response = requests.get(http_server.base_url + '/headers', headers={'Foo': 'foo', 'Bar': 'bar'})
    response.raise_for_status()

    headers = CaseInsensitiveDict()
    for line in response.text.splitlines():
        name, value = line.split(':', 1)
        headers[name] = value.strip()

    assert headers['Foo'] == 'foo'
    assert headers['Bar'] == 'bar'


def test_get_redirect_nofollow(http_server):
    response = requests.get(http_server.base_url + '/redirect', allow_redirects=False)
    response.raise_for_status()

    assert response.request.url.endswith('/redirect')
    assert response.url.endswith('/redirect')
    assert response.status_code == 302
    assert response.text == 'Redirecting...\n'


def test_get_redirect(http_server):
    with pytest.raises(requests.TooManyRedirects) as e:
        requests.get(http_server.base_url + '/redirect')

    # max_redirects defaults to 30
    assert http_server.last_url.path == '/redirect30'


def test_get_json(http_server):
    response = requests.get(http_server.base_url + '/json')
    response.raise_for_status()

    assert response.json() == {'Hello': 'World'}


def test_get_iter_content(http_server):
    response = requests.get(http_server.base_url + '/hello')
    response.raise_for_status()

    assert response.content == b'Hello\nWorld\n'
    it = response.iter_content(5, decode_unicode=True)
    assert next(it) == 'Hello'
    assert next(it) == '\nWorl'
    assert next(it) == 'd\n'
    with pytest.raises(StopIteration):
        assert next(it)


@pytest.mark.parametrize('delimiter', [None, '\n'])
def test_get_iter_lines(delimiter, http_server):
    response = requests.get(http_server.base_url + '/hello')
    response.raise_for_status()

    assert response.content == b'Hello\nWorld\n'
    it = response.iter_lines(5, decode_unicode=True, delimiter=delimiter)
    assert next(it) == 'Hello'
    assert next(it) == 'World'
    if delimiter is not None:
        assert next(it) == ''
    with pytest.raises(StopIteration):
        assert next(it)


TEST_COOKIEJAR = cookies.RequestsCookieJar()
TEST_COOKIEJAR.update({'a': 'Fizz', 'b': 'Bazz'})


@pytest.mark.parametrize('cookies_', [{'x': 'Foo', 'y': 'Bar'}, TEST_COOKIEJAR])
def test_cookies(http_server, cookies_):
    response = requests.get(http_server.base_url + '/cookies', cookies=cookies_)

    cookiejar = cookies.RequestsCookieJar()
    cookiejar.update(cookies_)
    assert response.text == '\n'.join('{}: {}'.format(*item) for item in cookiejar.items())


# Timeouts should go last, because '/slow' hangs the HTTP server
@pytest.mark.parametrize('timeout', [0.1, (None, 0.1)])
def test_get_timeout(http_server, timeout):
    with pytest.raises(requests.Timeout):
        requests.get(http_server.base_url + '/slow', timeout=timeout)


@pytest.mark.skipif('urllib3' not in sys.modules, reason='urllib3 not available')
@pytest.mark.parametrize('timeout', [Timeout(read=0.1), Timeout(total=0.1)])
def test_get_timeout_urllib3(http_server, timeout):
    with pytest.raises(requests.Timeout):
        requests.get(http_server.base_url + '/slow', timeout=timeout)


@pytest.mark.parametrize('timeout', [0.1, (0.1, None), Timeout(connect=0.1), Timeout(total=0.1)])
def test_get_connect_timeout(http_server, timeout):
    with pytest.raises(requests.Timeout):
        # Use RFC-5737 TEST-NET-1 address since it should always be unreachable
        requests.get('http://192.0.2.1', timeout=timeout)
        requests.get(http_server.base_url + '/slow', timeout=timeout)


@pytest.mark.skipif('urllib3' not in sys.modules, reason='urllib3 not available')
@pytest.mark.parametrize('timeout', [Timeout(connect=0.1), Timeout(total=0.1)])
def test_get_connect_timeout_urllib3(http_server, timeout):
    with pytest.raises(requests.Timeout):
        # Use RFC-5737 TEST-NET-1 address since it should always be unreachable
        requests.get('http://192.0.2.1', timeout=timeout)
        requests.get(http_server.base_url + '/slow', timeout=timeout)



from collections import abc
from typing import *

import pycurl

from pycurl_requests.models import Request, PreparedRequest, Response, DEFAULT_REDIRECT_LIMIT
from pycurl_requests import structures
from pycurl_requests import _pycurl


# Stubbed out for Requests tests
class SessionRedirectMixin:
    pass


class Session:
    def __init__(self, *, curl: pycurl.Curl = None):
        self.auth = None
        self.cert = None
        self.cookies = None
        self.headers = structures.CaseInsensitiveDict()
        self.hooks = NotImplemented
        self.max_redirects = DEFAULT_REDIRECT_LIMIT
        self.params = {}
        self.proxies = {}
        self.stream = False
        self.trust_env = True
        self.verify = True

        self.curl = curl or pycurl.Curl()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        if self.curl:
            self.curl.close()

        self.curl = None

    def get(self, url, params=None, **kwargs) -> Response:
        return self.request('GET', url, params=params, **kwargs)

    def head(self, url, **kwargs) -> Response:
        return self.request('HEAD', url, **kwargs)

    def options(self, url, **kwargs) -> Response:
        return self.request('OPTIONS', url, **kwargs)

    def post(self, url, data=None, json=None, **kwargs) -> Response:
        return self.request('POST', url, data=data, json=json, **kwargs)

    def put(self, url, data=None, **kwargs) -> Response:
        return self.request('PUT', url, data=data, **kwargs)

    def patch(self, url, data=None, **kwargs) -> Response:
        return self.request('PATCH', url, data=data, **kwargs)

    def delete(self, url, **kwargs) -> Response:
        return self.request('DELETE', url, **kwargs)

    def request(self, method, url,
                params=None, data=None, headers=None, cookies=None, files=None,
                auth=None, timeout=None, allow_redirects=True, proxies=None,
                hooks=None, stream=None, verify=None, cert=None, json=None) -> Response:
        request = Request(method, url,
                          params=params,
                          data=data,
                          json=json,
                          headers=headers,
                          cookies=cookies,
                          files=files,
                          auth=auth,
                          hooks=hooks)

        prepared = self.prepare_request(request)

        settings = dict(curl=self.curl, timeout=timeout, allow_redirects=allow_redirects,
                        max_redirects=self.max_redirects)
        settings.update(self.merge_environment_settings(prepared.url, proxies, stream, verify, cert))

        return self.send(prepared, **settings)

    def get_adapter(self, url) -> NotImplemented:
        raise NotImplementedError

    def get_redirect_target(self, resp: Response) -> Optional[str]:
        raise NotImplementedError

    def merge_environment_settings(self, url, proxies, stream, verify, cert) -> dict:
        # TODO: Read settings from environment
        return {}

    def mount(self, prefix, adapter):
        raise NotImplementedError

    def prepare_request(self, request: Request) -> PreparedRequest:
        prepared = PreparedRequest()

        headers = structures.CaseInsensitiveDict()
        for name, value in self.headers.items():
            headers[name] = value
        for name, value in (request.headers or {}).items():
            headers[name] = value

        prepared.prepare(
            method=request.method,
            url=request.url,
            headers=headers,
            files=request.files,
            data=request.data,
            json=request.json,
            params=_merge_params(self.params, request.params),
            auth=NotImplemented,  # TODO: Merge request with Session
            cookies=_merge_params(self.cookies, request.cookies),
            hooks=NotImplemented)  # TODO: Merge request with Session

        return prepared

    def rebuild_auth(self, prepared_request, response):
        raise NotImplementedError

    def rebuild_method(self, prepared_request, response):
        raise NotImplementedError

    def rebuild_proxies(self, prepared_request, proxies) -> dict:
        raise NotImplementedError

    def resolve_redirects(self, resp: Response, req: PreparedRequest, stream=False, timeout=None, verify=True, cert=None,
                          proxies=None, yield_requests=False, **adapter_kwargs) -> Generator:
        raise NotImplementedError

    def send(self, request: PreparedRequest, **kwargs):
        return _pycurl.send(request, **kwargs)

    def should_strip_auth(self, old_url, new_url):
        raise NotImplementedError


def _merge_params(current, new):
    """Merge parameters dictionary"""
    if not new:
        return current

    if not current:
        return new

    current = current.copy()
    current.update(new)

    return current


def session():
    """
    Create a Session

    .. deprecated:: 1.0.0
        Use :class:`~pycurl_requests.sessions.Session` instead.
    """
    return Session()



import datetime
import io
from io import BytesIO
import logging

import pycurl

from pycurl_requests import exceptions
from pycurl_requests import models
from pycurl_requests import structures

try:
    from urllib3.util.timeout import Timeout
except ImportError:
    # Timeout not supported
    Timeout = None


# For DEBUGFUNCTION callback
CURLINFO_TEXT = 0
CURLINFO_HEADER_IN = 1
CURLINFO_HEADER_OUT = 2

# Loggers
LOGGER = logging.getLogger('curl')
LOGGER_TEXT = LOGGER.getChild('text')
LOGGER_HEADER_IN = LOGGER.getChild('header_in')
LOGGER_HEADER_OUT = LOGGER.getChild('header_out')
DEBUGFUNCTION_LOGGERS = {LOGGER_TEXT, LOGGER_HEADER_IN, LOGGER_HEADER_OUT}

VERSION_INFO = pycurl.version_info()


class Request:
    def __init__(self, prepared, *, curl=None, timeout=None, allow_redirects=True, max_redirects=-1):
        self.prepared = prepared
        self.curl = curl or pycurl.Curl()
        self.timeout = timeout
        self.allow_redirects = allow_redirects
        self.max_redirects = max_redirects

        if timeout is not None:
            if isinstance(timeout, (int, float)):
                self.connect_timeout, self.read_timeout = timeout, timeout
            elif isinstance(timeout, Timeout):
                timeout.start_connect()
                self.connect_timeout = (0 if timeout.connect_timeout is Timeout.DEFAULT_TIMEOUT
                                        else timeout.connect_timeout)
                self.read_timeout = (0 if timeout.read_timeout is Timeout.DEFAULT_TIMEOUT
                                       else timeout.read_timeout)
            else:
                self.connect_timeout, self.read_timeout = timeout
        else:
            self.connect_timeout, self.read_timeout = (None, None)

        self.response_buffer = BytesIO()
        self.reason = None
        self.headers = structures.CaseInsensitiveDict()
        self.reset_headers = False

    def header_function(self, line: bytes):
        if self.reset_headers:
            self.headers = structures.CaseInsensitiveDict()
            self.reset_headers = False

        try:
            # Some servers return UTF-8 status
            line = line.decode('utf-8')
        except UnicodeDecodeError:
            # Fall back to latin-1
            line = line.decode('iso-8859-1')

        if self.reason is None:
            _, _, reason = line.split(' ', 2)
            self.reason = reason.strip()
            return

        if line == '\r\n':
            self.reset_headers = True
            return
        elif ':' not in line:
            return

        name, value = line.split(':', 1)
        self.headers[name] = value.strip()

    def send(self):
        try:
            # Avoid urlparse/urlsplit as they only support RFC 3986 compatible URLs
            scheme, _ = self.prepared.url.split(':', 1)
        except ValueError:
            raise exceptions.MissingSchema('Missing scheme for {!r}'.format(self.prepared.url))

        supported_protocols = VERSION_INFO[8]
        if scheme.lower() not in supported_protocols:
            raise exceptions.InvalidSchema('Unsupported scheme for {!r}'.format(self.prepared.url))

        # Request
        self.curl.setopt(pycurl.URL, self.prepared.url)

        if self.prepared.method:
            self.curl.setopt(pycurl.CUSTOMREQUEST, self.prepared.method)

        if self.prepared.method == 'HEAD':
            self.curl.setopt(pycurl.NOBODY, 1)

        self.curl.setopt(pycurl.HTTPHEADER, ['{}: {}'.format(n, v) for n, v in self.prepared.headers.items()])

        if self.prepared.body is not None:
            if isinstance(self.prepared.body, str):
                body = io.BytesIO(self.prepared.body.encode('iso-8859-1'))
            elif isinstance(self.prepared.body, bytes):
                body = io.BytesIO(self.prepared.body)
            else:
                body = self.prepared.body

            self.curl.setopt(pycurl.UPLOAD, 1)
            self.curl.setopt(pycurl.READDATA, body)

        content_length = self.prepared.headers.get('Content-Length')
        if content_length is not None:
            self.curl.setopt(pycurl.INFILESIZE_LARGE, int(content_length))

        # Response
        self.curl.setopt(pycurl.HEADERFUNCTION, self.header_function)
        self.curl.setopt(pycurl.WRITEDATA, self.response_buffer)

        # Options
        if self.connect_timeout is not None:
            timeout = int(self.connect_timeout * 1000)
            self.curl.setopt(pycurl.CONNECTTIMEOUT_MS, timeout)

        if self.read_timeout is not None:
            timeout = int(self.read_timeout * 1000)
            self.curl.setopt(pycurl.TIMEOUT_MS, timeout)

        if self.allow_redirects:
            self.curl.setopt(pycurl.FOLLOWLOCATION, 1)
            self.curl.setopt(pycurl.POSTREDIR, pycurl.REDIR_POST_ALL)
            self.curl.setopt(pycurl.MAXREDIRS, self.max_redirects)

        # Logging
        if any((l.isEnabledFor(logging.DEBUG) for l in DEBUGFUNCTION_LOGGERS)):
            self.curl.setopt(pycurl.VERBOSE, 1)
            self.curl.setopt(pycurl.DEBUGFUNCTION, debug_function)

        return self.perform()

    def perform(self):
        try:
            start_time = datetime.datetime.now(tz=datetime.timezone.utc)
            try:
                self.curl.perform()
            finally:
                end_time = datetime.datetime.now(tz=datetime.timezone.utc)
                self.prepared.url = self.curl.getinfo(pycurl.EFFECTIVE_URL)
                self.response_buffer.seek(0)
                response = self.build_response(elapsed=end_time - start_time)
        except pycurl.error as e:
            raise exceptions.RequestException.from_pycurl_error(
                e, request=self.prepared, response=response) from e

        return response

    def build_response(self, elapsed=None):
        status_code = self.curl.getinfo(pycurl.RESPONSE_CODE)
        if not status_code:
            return None

        response = models.Response()
        response.request = self.prepared
        response.elapsed = elapsed
        response.status_code = status_code
        response.reason = self.reason
        response.headers = self.headers
        response.encoding = self.headers.get_content_charset()
        response.url = self.prepared.url
        response.raw = self.response_buffer

        return response


def debug_function(infotype: int, message: bytes):
    """cURL `DEBUGFUNCTION` that writes to logger"""
    if infotype > CURLINFO_HEADER_OUT:
        # Ignore data messages
        return

    message = message.decode('utf-8', 'replace')

    if infotype == CURLINFO_TEXT:
        LOGGER_TEXT.debug(message.rstrip())
    elif infotype == CURLINFO_HEADER_IN:
        for line in message.splitlines():
            LOGGER_HEADER_IN.debug(line)
    elif infotype == CURLINFO_HEADER_OUT:
        for line in message.splitlines():
            LOGGER_HEADER_OUT.debug(line)


def send(*args, **kwargs):
    """Helper for making a Request and sending it."""
    return Request(*args, **kwargs).send()



import http.client


class CaseInsensitiveDict(http.client.HTTPMessage):
    """For compatibility with Requests."""
    def __init__(self, value=None, **kwargs):
        super(CaseInsensitiveDict, self).__init__()

        value = dict(value, **kwargs) if value is not None else kwargs
        for k, v in value.items():
            self[k] = v



"""
Test helper utilities.
"""

import json
import threading
import time
from http import cookies
from http.server import HTTPServer, BaseHTTPRequestHandler
from urllib.parse import urlsplit, parse_qsl

import pytest

from pycurl_requests import requests

__all__ = ['IS_PYCURL_REQUESTS', 'http_server']

#: Is this _really_ PyCurl-Requests?
#: Should be used when testing for PyCurl-Requests extensions.
IS_PYCURL_REQUESTS = requests.__name__ == 'pycurl_requests'


@pytest.fixture(scope='module')
def http_server():
    httpd = HTTPServer(('127.0.0.1', 0), HTTPRequestHandler)
    httpd.base_url = 'http://{}:{}'.format(*httpd.server_address)
    thread = threading.Thread(target=httpd.serve_forever)
    thread.start()
    try:
        yield httpd
    finally:
        httpd.shutdown()
        thread.join()


class HTTPRequestHandler(BaseHTTPRequestHandler):
    def log_message(self, format, *args):
        # Mute HTTP logging
        pass

    def do_GET(self):
        # Remember last requested URL
        self.server.last_url = self.url

        if self.url.path.startswith('/redirect'):
            self.do_GET_redirect()
        else:
            getattr(self, f'do_GET_{self.url.path[1:]}', self.do_HTTP_404)()

    def do_GET_hello(self):
        self.response('Hello\nWorld\n')

    def do_GET_params(self):
        self.response('\n'.join(('{}: {}'.format(n, v) for n, v in parse_qsl(self.url.query))))

    def do_GET_headers(self):
        self.response('\n'.join(('{}: {}'.format(n, v) for n, v in self.headers.items())))

    def do_GET_redirect(self):
        n = int(self.url.path[9:]) + 1 if len(self.url.path) > 9 else 1
        self.response('Redirecting...\n', (302, 'Found'), headers={'Location': f'/redirect{n}'})

    def do_GET_json(self):
        self.response(json.dumps({'Hello': 'World'}), content_type='application/json')

    def do_GET_slow(self):
        time.sleep(2)
        self.response('zZzZ\n')

    def do_GET_cookies(self):
        if 'Cookie' not in self.headers:
            self.send_error(400, 'No `Cookie` header sent')
            return

        cookie = cookies.SimpleCookie()
        cookie.load(self.headers['Cookie'])
        self.response('\n'.join(('{}: {}'.format(n, v.coded_value) for n, v in cookie.items())))

    def do_HTTP_404(self):
        self.send_error(404, 'Not Found')

    @property
    def url(self):
        if not hasattr(self, '_url'):
            # For some annoying reason this can't be set in `__init__`
            self._url = urlsplit(self.path)

        return self._url

    def response(self, body, status=None, content_type=None, headers=None):
        body = body.encode('utf-8') if isinstance(body, str) else body
        status = status or (200, 'OK')
        content_type = content_type or 'text/html; charset=UTF-8'
        headers = headers or {}
        self.send_response(*status)
        self.send_header('Content-Type', content_type)
        self.send_header('Content-Length', len(body))
        for key, value in headers.items():
            self.send_header(key, value)
        self.end_headers()

        self.wfile.write(body)



from pycurl_requests import sessions


def request(method, url, **kwargs):
    with sessions.Session() as session:
        return session.request(method, url, **kwargs)


def head(url, **kwargs):
    with sessions.Session() as session:
        return session.head(url, **kwargs)


def get(url, params=None, **kwargs):
    with sessions.Session() as session:
        return session.get(url, params=params, **kwargs)


def post(url, data=None, json=None, **kwargs):
    with sessions.Session() as session:
        return session.post(url, data=data, json=json, **kwargs)


def put(url, data=None, **kwargs):
    with sessions.Session() as session:
        return session.put(url, data=data, **kwargs)


def patch(url, data=None, **kwargs):
    with sessions.Session() as session:
        return session.patch(url, data=data, **kwargs)


def delete(url, params=None, **kwargs):
    with sessions.Session() as session:
        return session.delete(url, params=params, **kwargs)



import pycurl
import pytest

from pycurl_requests import requests
from pycurl_requests.tests.utils import *


def test_connecterror_refused():
    with pytest.raises(requests.ConnectionError) as e:
        requests.get('http://127.0.0.1:9')

    exception = e.value  # type: requests.ConnectionError

    assert exception.response is None
    assert exception.request.url == 'http://127.0.0.1:9/'

    if IS_PYCURL_REQUESTS:
        assert isinstance(exception.__cause__, pycurl.error)
        assert exception.curl_code == pycurl.E_COULDNT_CONNECT
        assert 'Connection refused' in exception.curl_message


def test_connecterror_resolve():
    with pytest.raises(requests.ConnectionError) as e:
        requests.get('http://nosuchdomain.example.com')

    exception = e.value  # type: requests.ConnectionError

    assert exception.response is None
    assert exception.request.url == 'http://nosuchdomain.example.com/'

    if IS_PYCURL_REQUESTS:
        assert isinstance(exception.__cause__, pycurl.error)
        assert exception.curl_code == pycurl.E_COULDNT_RESOLVE_HOST
        assert 'Could not resolve host' in exception.curl_message


def test_toomanyredirects(http_server):
    with pytest.raises(requests.TooManyRedirects) as e:
        with requests.Session() as s:
            s.max_redirects = 0
            s.get(http_server.base_url + '/redirect')

    exception = e.value  # type: requests.TooManyRedirects

    assert exception.response.status_code == 302
    assert exception.response.url == http_server.base_url + '/redirect'
    assert exception.request.url == http_server.base_url + '/redirect'

    if IS_PYCURL_REQUESTS:
        assert isinstance(exception.__cause__, pycurl.error)
        assert exception.curl_code == pycurl.E_TOO_MANY_REDIRECTS
        assert 'Maximum (0) redirects followed' in exception.curl_message



"""
Tests for Requests Sessions.

Note: We don't extensively test the various methods as they are exercised by
the tests for the tests for `api` (which itself is just a thin wrapper around a
single-use Session).
"""

import pycurl
import pytest

from pycurl_requests import requests
from pycurl_requests import cookies
from pycurl_requests.tests.utils import *  # Used for fixtures


def test_session():
    s = requests.Session()
    try:
        if IS_PYCURL_REQUESTS:
            assert isinstance(s.curl, pycurl.Curl)
    finally:
        s.close()

    if IS_PYCURL_REQUESTS:
        assert s.curl is None


def test_session_contextmanager():
    with requests.Session() as s:
        if IS_PYCURL_REQUESTS:
            assert isinstance(s.curl, pycurl.Curl)

    if IS_PYCURL_REQUESTS:
        assert s.curl is None


def test_session_max_redirects(http_server):
    with pytest.raises(requests.TooManyRedirects):
        with requests.Session() as s:
            s.max_redirects = 3
            s.get(http_server.base_url + '/redirect')

    assert http_server.last_url.path == '/redirect3'


def test_session_parameters(http_server):
    with requests.Session() as s:
        s.params = {'a': 1, 'b': 2}
        response = s.get(http_server.base_url + '/params', params={'b': 3, 'c': 4})

    assert response.text == 'a: 1\nb: 3\nc: 4'


COOKIEJAR1 = cookies.RequestsCookieJar()
COOKIEJAR1.update({'a': 'Fizz', 'b': 'Bazz'})

COOKIEJAR2 = cookies.RequestsCookieJar()
COOKIEJAR2.update({'b': 'Buzz', 'c': 'Boo'})


@pytest.mark.parametrize('cookies_', [{'b': 'Buzz', 'c': 'Boo'}, COOKIEJAR2])
def test_session_cookies(http_server, cookies_):
    with requests.Session() as s:
        s.cookies = COOKIEJAR1
        response = s.get(http_server.base_url + '/cookies', cookies=cookies_)

    assert response.text == 'a: Fizz\nb: Buzz\nc: Boo'



#!/usr/bin/env python3
# A helper for executing Request's unit-test suite
#
# See https://github.com/psf/requests/tree/master/tests

import os
import sys
import runpy
import pkg_resources

import requests.compat
import pycurl_requests

REQUIRED_PACKAGES = ['pytest-httpbin', 'pytest-mock']


if __name__ == '__main__':
    if not os.path.exists('Pipfile'):
        print('ERROR: Requests\' unit tests expect to run from the top-level of the repository',
              file=sys.stderr)
        sys.exit(1)

    try:
        for package in REQUIRED_PACKAGES:
            pkg_resources.get_distribution(package)
    except pkg_resources.DistributionNotFound:
        print('ERROR: Requests\' unit tests require additional fixtures:', file=sys.stderr)
        print('  pip install --user {}'.format(' '.join(REQUIRED_PACKAGES)), file=sys.stderr)
        sys.exit(1)

    pycurl_requests.patch_requests()

    # Requests' tests use compatibility functions
    sys.modules['requests.compat'] = requests.compat

    runpy.run_module('pytest', run_name='__main__')



import os

from .api import request, get, head, post, patch, put, delete
from .exceptions import RequestException, ConnectionError, HTTPError, URLRequired, TooManyRedirects, ConnectTimeout, ReadTimeout, Timeout
from .models import Request, Response, PreparedRequest
from .sessions import Session, session

# If PYCURL_REQUESTS is set to a non-null value, then this module imports
# Requests instead of PyCurl-Requests.
#
# This is used in unit-tests to allow checking of Requests compatibility.
if os.getenv('PYCURLREQUESTS_REQUESTS', None):
    import requests
else:
    requests = __import__(__name__)


def patch_requests():
    """Patch Requests library with PycURL Requests"""
    import sys

    for module in ('requests', 'requests.api', 'requests.cookies', 'requests.exceptions',
                   'requests.models', 'requests.sessions', 'requests.structures'):
        sys.modules[module] = sys.modules[module.replace('requests', __name__, 1)]



"""Shim for running existing Requests-based script using PycURL-Requests."""
import argparse
import runpy
import sys

import pycurl_requests


def main():
    parser = argparse.ArgumentParser(description='a Requests-compatible interface for PycURL')
    parser.add_argument('-m', dest='is_module', action='store_true',
                        help='run library module as script')
    parser.add_argument('name', help='script or module run')
    parser.add_argument('arg', nargs=argparse.REMAINDER, help='script arguments')
    args = parser.parse_args()

    # Override requests
    pycurl_requests.patch_requests()

    # Run script
    sys.argv[1:] = args.arg
    if args.is_module:
        runpy.run_module(args.name, run_name='__main__', alter_sys=True)
    else:
        runpy.run_path(args.name, run_name='__main__')
    sys.exit()


if __name__ == '__main__':
    main()
